﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DataAccess.Core.Execute;
using DataAccess.Core.Interfaces;

namespace DataAccess.Core
{
    /// <summary>
    /// Context to use for transactions
    /// </summary>
    public class TransactionContext : IDisposable
    {
        /// <summary>
        /// Operations ran here will happen in the context of a transaction
        /// </summary>
        public IDataStore Instance { get; set; }

        /// <summary>
        /// Information about the transaction
        /// </summary>
        public TransactionInfo TransactionInfo { get; set; }

        /// <summary>
        /// If true, on dispose the transaction will commit else rollback
        /// </summary>
        public bool CommitOnDispose { get; set; }

        /// <summary>
        /// Initializes a new instance of the <see cref="TransactionContext" /> class.
        /// </summary>
        /// <param name="store">The datastore to use.</param>
        public TransactionContext(IDataStore store)
        {
            CommitOnDispose = true;
            
            Instance = store.GetNewInstance();
            StartNewTransaction();
        }

        /// <summary>
        /// Rollbacks a transaction
        /// </summary>
        public void Rollback()
        {
            if (TransactionInfo != null)
            {
                TransactionInfo.Transaction.Rollback();
                TransactionInfo.Dispose();
                TransactionInfo = null;
            }
        }

        /// <summary>
        /// Commits a transaction
        /// </summary>
        public void Commit()
        {
            if (TransactionInfo != null)
            {
                TransactionInfo.Transaction.Commit();
                TransactionInfo.Dispose();
                TransactionInfo = null;
            }
        }

        /// <summary>
        /// Starts a new transaction.
        /// </summary>
        public void StartNewTransaction()
        {
            TransactionInfo = new TransactionInfo();
            TransactionInfo.Connection = Instance.Connection.GetConnection();
            TransactionInfo.Connection.Open();
            TransactionInfo.Transaction = TransactionInfo.Connection.BeginTransaction();
            Instance.ExecuteCommands = new TransactionCommandExecutor(TransactionInfo, Instance.ExecuteCommands);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            if (TransactionInfo != null)
            {
               if( CommitOnDispose)
                   Commit();
                else
                   Rollback();
            }
        }
    }
}
