﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DataAccess.Core;
using DataAccess.Core.Interfaces;
using System.Data;
using DataAccess.Core.Data;
using System.Collections;
using System.Reflection;
using System.IO;
using System.Data.SqlServerCe;

namespace DataAccess.SqlCompact
{
    /// <summary>
    /// Generates various types of data store Commands, appropriate for Sql Server
    /// </summary>
    public class SqlCompactCommandGenerator : DatabaseCommandGenerator
    {
        static SqlCompactCommandGenerator()
        {
            Assembly asmb = Assembly.GetExecutingAssembly();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SqlServerCompactCommandGenerator"/> class.
        /// </summary>
        public SqlCompactCommandGenerator()
        {
        }

        /// <summary>
        /// Returns a command for inserting one object
        /// </summary>
        /// <param name="item">The object to insert</param>
        /// <returns></returns>
        public override IDbCommand GetInsertCommand(object item)
        {
            TypeInfo ti = base.DataStore.TypeInformationParser.GetTypeInfo(item.GetType());
            if (ti.PrimaryKeys.Count == 1)
            {
                if (ti.PrimaryKeys[0].PropertyType == typeof(string))
                {
                    IDbCommand cmd = base.GetInsertCommand(item);
                    cmd.CommandText = cmd.CommandText.Replace("VALUES", "output inserted.* VALUES");
                    return cmd;                    
                }
                else
                {//this will only work with int keys
                    IDbCommand cmd = base.GetInsertCommand(item);
                    cmd.CommandText = cmd.CommandText.Replace(";", "; select SCOPE_IDENTITY() as " + ti.PrimaryKeys[0].EscapedFieldName);
                    return cmd;
                }                
            }
            else
            {
                return base.GetInsertCommand(item);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="items"></param>
        /// <returns></returns>
        public override IDbCommand GetInsertCommand(IList items)
        {//this can cause issues with triggers.
            IDbCommand cmd = base.GetInsertCommand(items);
            cmd.CommandText = cmd.CommandText.Replace("VALUES", "output inserted.* VALUES");
            return cmd;
        }

        /// <summary>
        /// Returns a command for creating a new table
        /// </summary>
        /// <param name="ti">The type to create a table for</param>
        /// <returns></returns>
        public override IDbCommand GetAddTableCommand(TypeInfo ti)
        {
            StringBuilder sb = new StringBuilder();
            StringBuilder pFields = new StringBuilder();
            StringBuilder fKey = new StringBuilder();
            SqlCeCommand cmd = new SqlCeCommand();

            sb.AppendFormat("CREATE TABLE {0} (", ResolveTableName(ti));
            for (int i = 0; i < ti.DataFields.Count; i++)
            {
                DataFieldInfo dfi = ti.DataFields[i];
                if (i > 0) sb.Append(",");

                if (dfi.PrimaryKey)
                {
                    if (pFields.Length > 0) pFields.Append(",");
                    pFields.Append(dfi.FieldName);

                    if (dfi.PropertyType == typeof(int) && ti.PrimaryKeys.Count == 1)
                        sb.AppendFormat("{0} {1} NOT NULL IDENTITY(1,1) ", dfi.EscapedFieldName, TranslateTypeToSql(dfi));
                    else
                        sb.AppendFormat("{0} {1} NOT NULL ", dfi.EscapedFieldName, TranslateTypeToSql(dfi));

                    if (dfi.PropertyType == typeof(string) && string.IsNullOrEmpty(dfi.DefaultValue)) dfi.DefaultValue = "newid()"; //if the user specifies a default value, dont override it
                }
                else
                {
                    sb.AppendFormat("{0} {1} NULL ", dfi.EscapedFieldName, TranslateTypeToSql(dfi));
                }

                if (dfi.PrimaryKeyType != null)
                    fKey.Append(GetForeignKeySql(dfi, ti, DataStore.TypeInformationParser.GetTypeInfo(dfi.PrimaryKeyType)));

                if (!string.IsNullOrEmpty(dfi.DefaultValue))
                    fKey.Append(GetDefaultValueSql(ti, dfi, cmd));
            }
            sb.Append(");");

            if (pFields.Length > 0)
            {
                sb.AppendFormat("GO; ALTER TABLE {0} ADD CONSTRAINT PK_{2}_{3} PRIMARY KEY ({1});",
                                ResolveTableName(ti), pFields.ToString(), ti.UnEscapedSchema, ti.UnescapedTableName);
            }
            sb.Append(fKey.ToString());

            cmd.CommandText = sb.ToString();
            return cmd;
        }

        private string GetDefaultValueSql(TypeInfo ti, DataFieldInfo dfi, SqlCeCommand cmd)
        {
            return string.Format("ALTER TABLE {0} ADD CONSTRAINT DF_{3}_{4}_{5} DEFAULT {2} FOR {1};", ResolveTableName(ti), dfi.FieldName, dfi.DefaultValue, ti.UnEscapedSchema, ti.UnescapedTableName, dfi.FieldName);
        }

        private string GetForeignKeySql(DataFieldInfo field, TypeInfo targetTable, TypeInfo pKeyTable)
        {
            StringBuilder sb = new StringBuilder("ALTER TABLE ");
            sb.AppendFormat("{0} ", ResolveTableName(targetTable));
            sb.AppendFormat("ADD CONSTRAINT FK_{0}_{1}_{2} ", targetTable.UnEscapedSchema, targetTable.UnescapedTableName, field.FieldName);
            sb.AppendFormat("FOREIGN KEY({0}) ", field.EscapedFieldName);
            sb.AppendFormat("REFERENCES {0}.{1} ({2}) ", pKeyTable.Schema, pKeyTable.TableName, pKeyTable.PrimaryKeys[0].EscapedFieldName);
            sb.AppendFormat("ON UPDATE {0} ON DELETE {0};", TranslateFkeyType(field.ForeignKeyType));

            return sb.ToString();
        }

        /// <summary>
        /// Returns a command for removing a column from a table
        /// </summary>
        /// <param name="type">The type to remove the column from</param>
        /// <param name="dfi">The column to remove</param>
        /// <returns></returns>
        public override IDbCommand GetRemoveColumnCommand(TypeInfo type, DataFieldInfo dfi)
        {
            SqlCeCommand scmd = new SqlCeCommand();
            scmd.CommandText = string.Format("ALTER TABLE {0} DROP COLUMN [{1}]", ResolveTableName(type), dfi.FieldName);
            return scmd;
        }

        /// <summary>
        /// Returns a command for adding a column to a table
        /// </summary>
        /// <param name="type">The type to add the column to</param>
        /// <param name="dfi">The column to add</param>
        /// <returns></returns>
        public override IDbCommand GetAddColumnCommnad(TypeInfo type, DataFieldInfo dfi)
        {
            if (dfi.PrimaryKey)
                throw new DataStoreException("Adding a primary key to an existing table is not supported");
            else
            {
                SqlCeCommand scmd = new SqlCeCommand();
                StringBuilder sb = new StringBuilder();
                sb.AppendFormat("ALTER TABLE {0} ADD {1} {2} NULL;", ResolveTableName(type), dfi.EscapedFieldName, TranslateTypeToSql(dfi));

                if (dfi.PrimaryKeyType != null)
                    sb.Append(GetForeignKeySql(dfi, type, DataStore.TypeInformationParser.GetTypeInfo(dfi.PrimaryKeyType)));

                if (!string.IsNullOrEmpty(dfi.DefaultValue))
                    sb.Append(GetDefaultValueSql(type, dfi, scmd));

                scmd.CommandText = sb.ToString();
                return scmd;
            }
        }        

        /// <summary>
        /// Changes types around
        /// </summary>
        /// <param name="dfi"></param>
        /// <returns></returns>
        public override string TranslateTypeToSql(DataFieldInfo dfi)
        {
            if (string.IsNullOrEmpty(dfi.DataFieldType)) // no user specified type, generate column type
            {
                if (dfi.PropertyType.Name.ToUpper() == "DATETIMEOFFSET")
                    return "DATETIMEOFFSET(7)";
                else if (dfi.PropertyType.Namespace.Equals("Microsoft.SqlServer.Types"))
                    return dfi.PropertyType.Name.Replace("Sql", "");
                else
                {


                    string toReturn = dfi.DataFieldType;
                    if (string.IsNullOrEmpty(toReturn)) // no user specified, generate column type
                    {
                        if (dfi.PrimaryKeyType != null)
                        { //use the primary keys field type instead					
                            toReturn = TranslateTypeToSql(DataStore.TypeInformationParser.GetTypeInfo(dfi.PrimaryKeyType).PrimaryKeys.First());
                        }
                        else
                        {
                            Type targetType = dfi.PropertyType;
                            if (targetType.IsGenericType && targetType.GetGenericTypeDefinition() == typeof(Nullable<>))
                                targetType = targetType.GetGenericArguments()[0];

                            if (targetType.IsEnum)
                                targetType = Enum.GetUnderlyingType(targetType);

                            string toChange = targetType.Name.ToUpper();
                            switch (toChange)
                            {
                                case "STRING":
                                    toReturn = "nvarchar(200)";
                                    break;
                                case "INT32":
                                    toReturn = "int";
                                    break;
                                case "INT64":
                                    toReturn = "bigint";
                                    break;
                                case "SINGLE":
                                case "DOUBLE":
                                    toReturn = "real";
                                    break;
                                case "BYTE[]":
                                    toReturn = "varbinary(MAX)";
                                    break;
                                case "BOOLEAN":
                                    toReturn = "bit";
                                    break;
                                case "DATETIME":
                                case "DATETIMEOFFSET":
                                    toReturn = "DATETIME";
                                    break;
                                case "TIMESPAN":
                                    toReturn = "time(7)";
                                    break;
                                case "CHAR":
                                    toReturn = "nvarchar(1)";
                                    break;

                            }
                        }
                    }
                    return toReturn;





                }
            }

            return dfi.DataFieldType;
        }

        /// <summary>
        /// Gets a command for changing a column type
        /// </summary>
        /// <param name="type">The type</param>
        /// <param name="dfi">The field</param>
        /// <param name="targetFieldType">The new column type</param>
        /// <returns></returns>
        public override IDbCommand GetModifyColumnCommand(TypeInfo type, DataFieldInfo dfi, string targetFieldType)
        {
            SqlCeCommand cmd = new SqlCeCommand();
            cmd.CommandText = string.Format("ALTER TABLE {0} ALTER COLUMN [{1}] {2}", ResolveTableName(type), dfi.FieldName, targetFieldType);
            return cmd;

        }

        /// <summary>
        /// /// Gets a table name for a type
        /// </summary>
        /// <param name="ti">The table name</param>
        /// <param name="EscapeTableName">ignored</param>
        /// <returns></returns>
        public override string ResolveTableName(TypeInfo ti, bool EscapeTableName)
        {
            return ResolveTableName(ti);
        }

        /// <summary>
        /// Gets a table name for a type
        /// </summary>
        /// <param name="ti">The type</param>
        /// <returns></returns>
        public override string ResolveTableName(TypeInfo ti)
        {
            if (!string.IsNullOrEmpty(ti.UnEscapedSchema))
                return string.Concat(ti.UnEscapedSchema, "_", ti.UnescapedTableName);
            else
                return ti.TableName;
        }

        public override string ResolveTableName(string schema, string table)
        {
            if (!string.IsNullOrEmpty(schema))
                return string.Concat(schema, "_", table);
            else
                return table;
        }

        public override IDbCommand GetAddSchemaCommand(TypeInfo ti)
        {
            return null;
        }
    }
}
