﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DataAccess.Core;
using DataAccess.Core.Interfaces;
using System.Data;
using DataAccess.Core.Data;
using System.Collections;
using System.Data.SqlClient;
using System.Reflection;
using System.IO;

namespace DataAccess.SqlServer
{
    /// <summary>
    /// Generates various types of data store Commands, appropriate for Sql Server
    /// </summary>
    public class SqlServerCommandGenerator : DatabaseCommandGenerator
    {
        private static string _createSchema;

        static SqlServerCommandGenerator()
        {
            Assembly asmb = Assembly.GetExecutingAssembly();
            _createSchema = asmb.LoadResource("DataAccess.SqlServer.Sql.CreateSchema.sql");
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SqlServerCommandGenerator"/> class.
        /// </summary>
        public SqlServerCommandGenerator()
        {
        }

        /// <summary>
        /// Returns a command for inserting one object
        /// </summary>
        /// <param name="item">The object to insert</param>
        /// <returns></returns>
        public override IDbCommand GetInsertCommand(object item)
        {
            TypeInfo ti = base.DataStore.TypeInformationParser.GetTypeInfo(item.GetType());
            if (ti.PrimaryKeys.Count == 1)
            {
                if (ti.PrimaryKeys[0].PropertyType == typeof(string))
                {
                    IDbCommand cmd = base.GetInsertCommand(item);
                    cmd.CommandText = cmd.CommandText.Replace("VALUES", "output inserted.* VALUES");
                    return cmd;                    
                }
                else
                {//this will only work with int keys
                    IDbCommand cmd = base.GetInsertCommand(item);
                    cmd.CommandText = cmd.CommandText.Replace(";", "; select SCOPE_IDENTITY() as " + ti.PrimaryKeys[0].EscapedFieldName);
                    return cmd;
                }                
            }
            else
            {
                return base.GetInsertCommand(item);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="items"></param>
        /// <returns></returns>
        public override IDbCommand GetInsertCommand(IList items)
        {//this can cause issues with triggers.
            IDbCommand cmd = base.GetInsertCommand(items);
            cmd.CommandText = cmd.CommandText.Replace("VALUES", "output inserted.* VALUES");
            return cmd;
        }

        /// <summary>
        /// Returns a command for creating a new table
        /// </summary>
        /// <param name="ti">The type to create a table for</param>
        /// <returns></returns>
        public override IDbCommand GetAddTableCommand(TypeInfo ti)
        {
            StringBuilder sb = new StringBuilder();
            StringBuilder pFields = new StringBuilder();
            StringBuilder fKey = new StringBuilder();
            SqlCommand cmd = new SqlCommand();

            sb.AppendFormat("CREATE TABLE {0}.{1} (", ti.Schema, ti.TableName);
            for (int i = 0; i < ti.DataFields.Count; i++)
            {
                DataFieldInfo dfi = ti.DataFields[i];
                if (i > 0) sb.Append(",");

                if (dfi.PrimaryKey)
                {
                    if (pFields.Length > 0) pFields.Append(",");
                    pFields.Append(dfi.FieldName);

                    if (dfi.PropertyType == typeof(int) && ti.PrimaryKeys.Count == 1)
                        sb.AppendFormat("{0} {1} NOT NULL IDENTITY(1,1) ", dfi.EscapedFieldName, TranslateTypeToSql(dfi));
                    else
                        sb.AppendFormat("{0} {1} NOT NULL ", dfi.EscapedFieldName, TranslateTypeToSql(dfi));

                    if (dfi.PropertyType == typeof(string) && string.IsNullOrEmpty(dfi.DefaultValue)) dfi.DefaultValue = "newid()"; //if the user specifies a default value, dont override it
                }
                else
                {
                    sb.AppendFormat("{0} {1} NULL ", dfi.EscapedFieldName, TranslateTypeToSql(dfi));
                }

                if (dfi.PrimaryKeyType != null)
                    fKey.Append(GetForeignKeySql(dfi, ti, DataStore.TypeInformationParser.GetTypeInfo(dfi.PrimaryKeyType)));

                if (!string.IsNullOrEmpty(dfi.DefaultValue))
                    fKey.Append(GetDefaultValueSql(ti, dfi, cmd));
            }
            sb.Append(") ON [PRIMARY];");

            if (pFields.Length > 0)
            {
                sb.AppendFormat("ALTER TABLE {0}.{1} ADD CONSTRAINT PK_{3}_{4} PRIMARY KEY CLUSTERED ({2}) WITH( STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) ON [PRIMARY];",
                                ti.Schema, ti.TableName, pFields.ToString(), ti.UnEscapedSchema, ti.UnescapedTableName);
            }
            sb.Append(fKey.ToString());

            cmd.CommandText = sb.ToString();
            return cmd;
        }

        private string GetDefaultValueSql(TypeInfo ti, DataFieldInfo dfi, SqlCommand cmd)
        {
            return string.Format("ALTER TABLE {0}.{1} ADD CONSTRAINT DF_{4}_{5}_{6} DEFAULT {3} FOR {2};", ti.Schema, ti.TableName, dfi.FieldName, dfi.DefaultValue, ti.UnEscapedSchema, ti.UnescapedTableName, dfi.FieldName);
        }

        private string GetForeignKeySql(DataFieldInfo field, TypeInfo targetTable, TypeInfo pKeyTable)
        {
            StringBuilder sb = new StringBuilder("ALTER TABLE ");
            sb.AppendFormat("{0}.{1} ", targetTable.Schema, targetTable.TableName);
            sb.AppendFormat("ADD CONSTRAINT FK_{0}_{1}_{2} ", targetTable.UnEscapedSchema, targetTable.UnescapedTableName, field.FieldName);
            sb.AppendFormat("FOREIGN KEY({0}) ", field.EscapedFieldName);
            sb.AppendFormat("REFERENCES {0}.{1} ({2}) ", pKeyTable.Schema, pKeyTable.TableName, pKeyTable.PrimaryKeys[0].EscapedFieldName);
            sb.AppendFormat("ON UPDATE {0} ON DELETE {0};", TranslateFkeyType(field.ForeignKeyType));

            return sb.ToString();
        }

        /// <summary>
        /// Returns a command for removing a column from a table
        /// </summary>
        /// <param name="type">The type to remove the column from</param>
        /// <param name="dfi">The column to remove</param>
        /// <returns></returns>
        public override IDbCommand GetRemoveColumnCommand(TypeInfo type, DataFieldInfo dfi)
        {
            SqlCommand scmd = new SqlCommand();
            scmd.CommandText = string.Format("ALTER TABLE {0}.{1} DROP COLUMN [{2}]", type.Schema, type.TableName, dfi.FieldName);
            return scmd;
        }

        /// <summary>
        /// Returns a command for adding a column to a table
        /// </summary>
        /// <param name="type">The type to add the column to</param>
        /// <param name="dfi">The column to add</param>
        /// <returns></returns>
        public override IDbCommand GetAddColumnCommnad(TypeInfo type, DataFieldInfo dfi)
        {
            if (dfi.PrimaryKey)
                throw new DataStoreException("Adding a primary key to an existing table is not supported");
            else
            {
                SqlCommand scmd = new SqlCommand();
                StringBuilder sb = new StringBuilder();
                sb.AppendFormat("ALTER TABLE {0}.{1} ADD {2} {3} NULL;", type.Schema, type.TableName, dfi.EscapedFieldName, TranslateTypeToSql(dfi));

                if (dfi.PrimaryKeyType != null)
                    sb.Append(GetForeignKeySql(dfi, type, DataStore.TypeInformationParser.GetTypeInfo(dfi.PrimaryKeyType)));

                if (!string.IsNullOrEmpty(dfi.DefaultValue))
                    sb.Append(GetDefaultValueSql(type, dfi, scmd));

                scmd.CommandText = sb.ToString();
                return scmd;
            }
        }        

        /// <summary>
        /// Changes types around
        /// </summary>
        /// <param name="dfi"></param>
        /// <returns></returns>
        public override string TranslateTypeToSql(DataFieldInfo dfi)
        {
            if (string.IsNullOrEmpty(dfi.DataFieldType)) // no user specified type, generate column type
            {
                if (dfi.PropertyType.Name.ToUpper() == "DATETIMEOFFSET")
                    return "DATETIMEOFFSET(7)";
                else if (dfi.PropertyType.Namespace.Equals("Microsoft.SqlServer.Types"))
                    return dfi.PropertyType.Name.Replace("Sql", "");
                else
                    return base.TranslateTypeToSql(dfi);
            }

            return dfi.DataFieldType;
        }

        /// <summary>
        /// Gets a command for changing a column type
        /// </summary>
        /// <param name="type">The type</param>
        /// <param name="dfi">The field</param>
        /// <param name="targetFieldType">The new column type</param>
        /// <returns></returns>
        public override IDbCommand GetModifyColumnCommand(TypeInfo type, DataFieldInfo dfi, string targetFieldType)
        {
            SqlCommand cmd = new SqlCommand();
            cmd.CommandText = string.Format("ALTER TABLE {0} ALTER COLUMN [{1}] {2}", ResolveTableName(type), dfi.FieldName, targetFieldType);
            return cmd;

        }

        /// <summary>
        /// /// Gets a table name for a type
        /// </summary>
        /// <param name="ti">The table name</param>
        /// <param name="EscapeTableName">ignored</param>
        /// <returns></returns>
        public override string ResolveTableName(TypeInfo ti, bool EscapeTableName)
        {
            return ResolveTableName(ti);
        }

        /// <summary>
        /// Gets a table name for a type
        /// </summary>
        /// <param name="ti">The type</param>
        /// <returns></returns>
        public override string ResolveTableName(TypeInfo ti)
        {
            if (!string.IsNullOrEmpty(ti.Schema))
                return string.Concat(ti.Schema, ".", ti.TableName);
            else
                return ti.TableName;
        }

        public override string ResolveTableName(string schema, string table)
        {
            if (!string.IsNullOrEmpty(schema))
                return string.Concat(schema, ".", table);
            else
                return table;
        }

        /// <summary>
        /// Returns a command that is appropriate for adding a schema for the object to go into
        /// </summary>
        /// <param name="ti"></param>
        /// <returns></returns>
        public override IDbCommand GetAddSchemaCommand(TypeInfo ti)
        {
            if (!ti.UnEscapedSchema.Equals("dbo", StringComparison.InvariantCultureIgnoreCase))
            {
                IDbCommand cmd = new SqlCommand();
                cmd.CommandText = string.Format(_createSchema, ti.UnEscapedSchema);
                return cmd;
            }
            else
            {
                return null;
            }
        }
    }
}
